__version__ = "0.20.0rc1"
__author__ = "Tulir Asokan <tulir@maunium.net>"
__all__ = [
    "api",
    "appservice",
    "bridge",
    "client",
    "crypto",
    "errors",
    "util",
    "types",
    "__optional_imports__",
]

from typing import TYPE_CHECKING

__optional_imports__ = TYPE_CHECKING
